/**********************************************************************
* spr.h
*
* The basic sprite support system prototypes and structures.
**********************************************************************
                    This file is part of

          STK -- The sprite toolkit -- version 1.1

              Copyright (C) Jari Karjala 1991

The sprite toolkit (STK) is a FreeWare toolkit for creating high
resolution sprite graphics with PCompatible hardware. This toolkit 
is provided as is without any warranty or such thing. See the file
COPYING for further information.

**********************************************************************
**********************************************************************/

#ifndef __SPR_H_
#define __SPR_H_

#include "grtypes.h"

extern int spr_pass_delay;
/**********************************************************************
* The delay after setvisualpage() before removing old objects in
* the funtion spr_next_pass(). 
* Many EGA/VGA cards need some milliseconds to switch pages, 
* the sprites flicker if this delay is too short.
*
* If someone has better solutions in dealing with this problem,
* I sure would like to know about them.
*
* Time is given in milliseconds and the default is 10 ms for EGA
* and 0 ms for Hercules (use spr_regulate_speed to balance the frame
* rate with different display adapters).
**********************************************************************/

#define spr_max_x gr_max_x
#define spr_max_y gr_max_y
/**********************************************************************
* The maximum coordinate values for current graphics adapter.
* Variables defined in module gr.c, and initialized in gr_start.
* (It is faster to use variables instead of getmaxx() and getmaxy());
**********************************************************************/

#ifndef __SPRITE_
typedef void *SPRITE;
#endif
/**********************************************************************
* The sprite type is private. The handle must be void pointer due
* to the Turbo C which does not allow "struct _sprite *SPRITE" without
* struct _sprite definition in the same file.
**********************************************************************/

void spr_initialize(int graphicsdriver);
/**********************************************************************
* Initialize the sprite system to the given display hardware.
* Supported graphicsdrivers: EGA (only two colors), EGAMONO and HERCMONO.
* 
* The visual page is set to 0.
*
* NOTE: This function must be called before any other sprite funtions
*       and the graphics mode must have been set before this call.
* 
* graphicsdriver  The BGI identifier for the graphics driver used.
**********************************************************************/

SPRITE spr_create(WORD w, WORD h, 
                  BITMAP pic, BITMAP mask, 
                  BYTE res, WORD ID);
/**********************************************************************
* Create a sprite from the given bitmaps.
*
* w,h   Width (must be < 512) and height (<256) of sprite in pixels
* pic   The picture bitmap for the sprite
* mask  The mask bitmap for the sprite
* res   The number of steps wanted per 8 bit interval in horizontal
*       direction (1,2,4,8). For example, the value 8 gives one
*       pixel resolution in X-direction.
* ID    The user supplied ID for the sprite (usually index into an 
*       array and/or sprite type identifier).
*
* Return: the newly created sprite or NULL if parameter error,
*           sprite too big or out-of-memory
**********************************************************************/

SPRITE spr_share(SPRITE spr, BYTE n);
/**********************************************************************
* Create a shared version of the given sprite. This allows at
* most n spr_copies which all share the shape data thus saving
* quite much memory.
*
* spr   The sprite to share
* n     The maximum number of shared copies
*
* Return: New SPRITE or NULL if error (out of memory, spr==NULL)
**********************************************************************/

SPRITE spr_copy(SPRITE spr, WORD id);
/**********************************************************************
* Create a copy of the given sprite. If the sprite was shared
* with spr_share then a shared copy is created. 
*
* spr   The sprite to share
* id    The ID for the new sprite
*
* Return: New SPRITE or NULL if error (out of memory, spr was NULL, 
*           no more shared copies)
**********************************************************************/

void spr_put(SPRITE spr, WORD x, WORD y);
/**********************************************************************
* Put the sprite into the given position in the display. 
* Actually the call has no effect on screen until spr_next_pass() is
* called next time and the sprite will disappear after the next
* spr_next_pass() call after that.
*
* NOTE: Cannot be called twice for the same sprite before spr_hide() or
*       spr_next_pass().
*
* spr   The sprite to put
* x     The X coordinate
* y     The Y coordinate
**********************************************************************/

void spr_hide(SPRITE spr);
/**********************************************************************
* Undo a spr_put for the sprite. This function only removes the sprite
* from internal display list since the spr_put did not actually put the 
* sprite into the screen.
*
* spr   The sprite to hide
**********************************************************************/

void spr_delete(SPRITE spr);
/**********************************************************************
* Delete the given sprite and release associated memory buffers.
* Also spr_hides the sprite.
*
* spr   The sprite to delete
**********************************************************************/

WORD spr_next_pass(void);
/**********************************************************************
* 1. Draw all sprites into the hidden screen page.
* 2. Make the hidden screen page visible.
* 3. Delete old sprites from the new hidden page.
*
* Return: The current visual page number
**********************************************************************/

void spr_regulate_speed(void);
/**********************************************************************
* This function tries to regulate the frame speed by delaying if
* we are doing more frames per second than we should. The target 
* speed is 1 frame per one clock tick (about 18 frames per second).
* This is a reasonable target speed for 286 machines.
* This function should be called after each spr_next_pass() call.
**********************************************************************/


/***** Information retrieving functions *****/

WORD spr_get_id(SPRITE spr);
/**********************************************************************
* Return the user supplied identifier of the sprite
**********************************************************************/

WORD spr_get_x(SPRITE spr);
/**********************************************************************
* Return the X coordinate of the sprite in pixels from top-left
**********************************************************************/

WORD spr_get_y(SPRITE spr);
/**********************************************************************
* Return the Y coordinate of the sprite in pixels from top-left
**********************************************************************/

WORD spr_get_width(SPRITE spr);
/**********************************************************************
* Return the width of the sprite in pixels
**********************************************************************/

WORD spr_get_height(SPRITE spr);
/**********************************************************************
* Return the height of the sprite in pixels
**********************************************************************/

#endif
